####################################################################################
####################################################################################
####                                                                            ####
####                        R code - MAGBLUP - Functions                        ####
####                              Flint-Dent Panel                              ####
####                                 S.Rio 2020                                 ####
####                                                                            ####
####  Associated publication (Rio et al. 2020):                                 ####
####    Accounting for group-specific allele effects and admixture in genomic   ####
####    predictions: theory and experimental evaluation in maize                ####
####                                                                            ####
####  Functions to be used in the main script:                                  ####
####    - generate_effects: simulate allele effects with various                ####
####      configurations depending on the QTL ancestry                          ####
####    - generate_pheno: simulate phenotypes using allele effects simulated    ####
####      using generate_effects function                                       ####
####    - GBLUP_RI_Var: function to compute GBLUP-RI covariance matrices        ####
####    - GBLUP_RAE_Var: function to compute GBLUP-RAE covariance matrices      ####
####    - MAGBLUP_RI_Var: Function to compute MAGBLUP-RI covariance matrices    ####
####    - MAGBLUP_RAE_Var: Function to compute MAGBLUP-RAE covariance matrices  ####
####    - SHO_sampling: Function to define CV sample according to SHO procedure ####
####    - BLUP_MM4LMM: Function to compute BLUPs using MM4LMM outputs           ####
####                                                                            ####
####################################################################################
####################################################################################


generate_effects <- function(sigma2_effects_list,f_D,f_F,nb_qtl){
  ##  Function to simulate allele effects with various configurations depending 
  ##  on the QTL ancestry with the following arguments:                
  ##    - sigma2_effects_list: list of variances of allele effects 
  ##      (sigma2_gamma, sigma2_delta_D and sigma2_delta_F) 
  ##    - f_D: Dent allele frequency
  ##    - f_F: Flint allele frequency 
  ##    - nb_qtl: number of QTLs to be sampled among SNPs

  ## Number of SNPs
  nb_SNP <- length(f_D)
  
  ## Sampling of QTLs
  QTL_sample_sim <- sort(sample(1:nb_SNP,nb_qtl))
  
  ## Sampling of allele effects
  sim_effects <- lapply(sigma2_effects_list,function(i){
    gamma_effect <- matrix(rnorm(2*nb_qtl,sd = sqrt(i[1])),nb_qtl,2)
    delta_D_effect <- matrix(rnorm(2*nb_qtl,sd = sqrt(i[2])),nb_qtl,2)
    delta_F_effect <- matrix(rnorm(2*nb_qtl,sd = sqrt(i[3])),nb_qtl,2)
    beta_0D <- gamma_effect[,2]+delta_D_effect[,2]
    beta_1D <- gamma_effect[,1]+delta_D_effect[,1]
    beta_0F <- gamma_effect[,2]+delta_F_effect[,2]
    beta_1F <- gamma_effect[,1]+delta_F_effect[,1]
    objet <- list(beta_0D,beta_1D,beta_0F,beta_1F)
    return(objet)
  })
  names(sim_effects) <- names(sigma2_effects_list)
  
  ## Reference variances of MAGBLUP-RI  
  sigma2_S <- sapply(sim_effects,function(i){
    mu_mD <- i[[1]]+f_D[QTL_sample_sim]*(i[[2]]-i[[1]])
    mu_mF <- i[[3]]+f_F[QTL_sample_sim]*(i[[4]]-i[[3]])
    res <- 1/(nb_qtl-1)*(nb_qtl*sum((mu_mD-mu_mF)^2)-sum(mu_mD-mu_mF)^2)
    return(res)
  })
  sigma2_GD <- sapply(sim_effects,function(i)
    sum(f_D[QTL_sample_sim]*(1-f_D[QTL_sample_sim])*(i[[2]]-i[[1]])^2))
  sigma2_GF <- sapply(sim_effects,function(i)
    sum(f_F[QTL_sample_sim]*(1-f_F[QTL_sample_sim])*(i[[4]]-i[[3]])^2))
  names(sigma2_S) <- names(sigma2_GD) <- names(sigma2_GF) <- names(sigma2_effects_list)
  
  ## Reference variances of MAGBLUP-RAE
  Sigma2_U <- sapply(sigma2_effects_list,function(i)i[1]*nb_qtl)
  Sigma2_UD <- sapply(sigma2_effects_list,function(i)i[2]*nb_qtl)
  Sigma2_UF <- sapply(sigma2_effects_list,function(i)i[3]*nb_qtl)
  names(Sigma2_U) <- names(Sigma2_UD) <- names(Sigma2_UF) <- names(sigma2_effects_list)
  
  return(list(effects=sim_effects,QTL_sample=QTL_sample_sim,
              var_MAGBLUP_RI=list(sigma2_S=sigma2_S,sigma2_GD=sigma2_GD,sigma2_GF=sigma2_GF),
              var_MAGBLUP_RAE=list(Sigma2_U=Sigma2_U,Sigma2_UD=Sigma2_UD,Sigma2_UF=Sigma2_UF)))
}


generate_pheno <- function(genotype,dent_ancestry,simulated_effects,QTL_sample_sim,h2){
  ##  Function to simulate phenotypes using allele effects simulated 
  ##  using generate_effects function with the following arguments:                
  ##    - genotype:  genotypic matrix 
  ##    - dent_ancestry: matrix indicating dent allele ancestry
  ##    - simulated_effects: simulated allele effects 
  ##    - QTL_sample_sim: sample of SNP declared as QTLs
  ##    - h2: heritability of simulated traits

  ## Parameters
  nb_ind <- nrow(genotype)
  nb_config <- length(simulated_effects)
  
  ## Genetic values
  A_sub <- dent_ancestry[,QTL_sample_sim]
  genotype_sub <- genotype[,QTL_sample_sim]
  TBV <- sapply(1:nb_config,function(i)(A_sub*genotype_sub)%*%(simulated_effects[[i]][[2]])+
                (A_sub*(1-genotype_sub))%*%simulated_effects[[i]][[1]]+
                ((1-A_sub)*genotype_sub)%*%(simulated_effects[[i]][[4]])+
                ((1-A_sub)*(1-genotype_sub))%*%simulated_effects[[i]][[3]])
  
  ## Phenotypes
  res <- sapply(1:nb_config,function(i)rnorm(nb_ind,sd=sqrt(var(TBV[,i])*(1/h2-1))))
  pheno <- sapply(1:nb_config,function(i)TBV[,i]+res[,i])
  
  rownames(TBV) <- rownames(pheno) <- rownames(genotype)
  colnames(TBV) <- colnames(pheno) <- names(simulated_effects)

  return(list(TBV=TBV,pheno=pheno))
}

GBLUP_RI_Var <- function(geno){
  ##  Function to compute GBLUP-RI covariance matrices:                
  ##    - geno:  genotypic matrix 
  
  ## Number of individuals
  nb_ind <- nrow(geno)
  
  ## Vector of allele frequencies
  freq <- colMeans(geno)

  ## Genotypic matrices centered by group-specific allele frequencies
  W <- (geno-tcrossprod(rep(1,nb_ind),freq))

  ## Kinship matrix (VanRaden 2008)
  K <- tcrossprod(W)/sum(freq*(1-freq))
  
  ## Identity matrix
  I <- diag(nb_ind)
  
  return(list(K,I))
}

GBLUP_RAE_Var <- function(geno){
  ##  Function to compute GBLUP-RAE covariance matrices:                
  ##    - geno:  genotypic matrix 

  ## Number of SNPs and individuals
  nb_SNP <- ncol(geno)
  nb_ind <- nrow(geno)
  
  ## Identity-by-state (IBS) matrix
  Phi <- (tcrossprod(geno)+tcrossprod(1-geno))/nb_SNP
  
  ## Identity matrix
  I <- diag(nb_ind)
  
  return(list(Phi,I))
}


MAGBLUP_RI_Var <- function(geno,ancestry){
  ##  Function to compute MAGBLUP-RI covariance matrices:                
  ##    - geno:  genotypic matrix 
  ##    - ancestry: matrix indicating allele ancestry
  
  ## Number of SNPs and individuals
  nb_SNP <- ncol(geno)
  nb_ind <- nrow(geno)
  
  ## Matrix of admixture proportions
  admixture  <-  rowMeans(ancestry)
  
  ## Vectors of group-specific allele frequencies
  freq_A <- colSums(ancestry*geno)/colSums(ancestry)
  freq_B <- colSums((1-ancestry)*geno)/colSums((1-ancestry))
  
  ## Genotypic matrices centered by group-specific allele frequencies
  W_A <- (geno-tcrossprod(rep(1,nb_ind),freq_A))
  W_B <- (geno-tcrossprod(rep(1,nb_ind),freq_B))
  
  ## Group-specific kinship matrices
  K_A <- tcrossprod(ancestry*W_A)/(tcrossprod(ancestry*tcrossprod(rep(1,nb_ind),freq_A),
                                              ancestry*tcrossprod(rep(1,nb_ind),1-freq_A)))
  K_A[which(K_A=="NaN")] <- 0
  K_B <- tcrossprod((1-ancestry)*W_B)/(tcrossprod((1-ancestry)*tcrossprod(rep(1,nb_ind),freq_B),
                                                  (1-ancestry)*tcrossprod(rep(1,nb_ind),1-freq_B)))
  K_B[which(K_B=="NaN")] <- 0
  
  ## Covriance matrix between ancestries
  Delta <- tcrossprod(ancestry)/nb_SNP-tcrossprod(admixture)
  
  ## Matrix of group-specific shared ancestries
  theta_A <- tcrossprod(ancestry)/nb_SNP
  theta_B <- tcrossprod(1-ancestry)/nb_SNP
  
  ## Identity matrix
  I <- diag(nb_ind)
  
  return(list(Delta,theta_A*K_A,theta_B*K_B,I))
}


MAGBLUP_RAE_Var <- function(geno,ancestry){
  ##  Function to compute MAGBLUP-RAE covariance matrices:                
  ##    - geno:  genotypic matrix 
  ##    - ancestry: matrix indicating allele ancestry
  
  ## Number of SNPs and individuals
  nb_SNP <- ncol(geno)
  nb_ind <- nrow(geno)
  
  ## Identity-by-state (IBS) matrix
  Phi <- (tcrossprod(geno)+tcrossprod(1-geno))/nb_SNP
  
  ## Identity-by-state (IBS) on shared ancestry in each group
  Phi_A <- (tcrossprod(ancestry*geno)+tcrossprod(ancestry*(1-geno)))/nb_SNP
  Phi_B <- (tcrossprod((1-ancestry)*geno)+tcrossprod((1-ancestry)*(1-geno)))/nb_SNP
  
  ## Identity matrix
  I <- diag(nb_ind)
  
  return(list(Phi,Phi_A,Phi_B,I))
}

SHO_sampling <- function(nb_TS,nb_PS,genetic_background){
  ##  Function to define CV sample according to SHO procedure with arguments:                
  ##    - nb_TS:  vector with number of individuals for each genetic background (D,F,A)
  ##    - nb_PS:  vector with number of individuals for each genetic background (D,F,A)
  ##    - genetic_background: data.frame with genetc background of individuals
  
  ## Number of individuals
  nb_ind <- nrow(genetic_background)
  
  ## Training and predicted set
  TS <- c(sample(c(1:nb_ind)[which(genetic_background$genetic_background=="Dent")],nb_TS[1]),
       sample(c(1:nb_ind)[which(genetic_background$genetic_background=="Flint")],nb_TS[2]),
       sample(c(1:nb_ind)[which(genetic_background$genetic_background=="Admixed")],nb_TS[3]))
  PS <- c(sample(c(1:nb_ind)[-union(which(genetic_background$genetic_background=="Flint"|
                                         genetic_background$genetic_background=="Admixed"),y = TS)],nb_PS[1]),
       sample(c(1:nb_ind)[-union(which(genetic_background$genetic_background=="Dent"|
                                         genetic_background$genetic_background=="Admixed"),y = TS)],nb_PS[2]),
       sample(c(1:nb_ind)[-union(which(genetic_background$genetic_background=="Dent"|
                                         genetic_background$genetic_background=="Flint"),y = TS)],nb_PS[3]))
  return(list(TS=sort(TS),PS=sort(PS)))
}

BLUP_MM4LMM <- function(model,Y,X,List_var,List_Z,Component){
  ##  Function to compute BLUPs using MM4LMM outputs:                
  ##    - model:  model object generated by MM4LMMM
  ##    - Y: vector of phenotypes
  ##    - X: design matrix for fixed effects
  ##    - List_var: List of covariance matrices
  ##    - List_Z: List of indicidence matrices
  ##    - Component: vector of indices of model components on which to compute BLUP
  
  inv_var_y <- solve(Reduce(lapply(1:length(model$NullModel$Sigma2),function(i)
    List_Z[[i]]%*%List_var[[i]]%*%t(List_Z[[i]])*model$NullModel$Sigma2[i]),f = "+"))
  BLUP_list <- lapply(Component,function(i)
    (List_var[[i]]*model$NullModel$Sigma2[i])%*%t(List_Z[[i]])%*%inv_var_y%*%
                     (Y-X%*%model$NullModel$Beta))
  return(BLUP_list)
}
